"use strict";

const Node_1 = require('./Node');
const Subject_1 = require('./Subject');
const Resource_1 = require('./Resource');
const util_1 = require('../util');
;
/**
  Class for building Resource/Subject class hierarchy based on schema
 */
class Graph {
    constructor(schema) {
        this.schema = schema;
        this.resources = Graph.buildResourceHierachy(schema.resources);
        this.subjects = Graph.buildSubjectHierachy(schema.subjects);
    }
    /**
     *  Build map of nodeName -> ResourceClass for gracl hierarchy
        NOTE: unfortunately we need to duplicate this function for subjects
              due to limitations of the way type parameters are handled
          See: https://github.com/Microsoft/TypeScript/issues/4890
     */
    static buildResourceHierachy(schemaNodes) {
        const nodeList = util_1.topologicalSort(schemaNodes),
              classGraph = new Map();
        const createClass = node => {
            const getParentsMethod = node.getParents || Node_1.Node.prototype.getParents;
            const getPermissionMethod = node.getPermission || Resource_1.Resource.prototype.getPermission;
            if (node.parent) {
                const ParentClass = classGraph.get(node.parent);
                classGraph.set(node.name, (_a = class extends ParentClass {
                    getParents() {
                        return getParentsMethod.call(this);
                    }

                    getPermission(subject) {
                        return getPermissionMethod.call(this, subject);
                    }
                }, _a.id = node.id || 'id', _a.parentId = node.parentId, _a.displayName = node.name, _a.repository = node.repository, _a.permissionPropertyKey = node.permissionProperty || 'permissions', _a));
            } else {
                classGraph.set(node.name, (_b = class extends Resource_1.Resource {
                    getPermission(subject) {
                        return getPermissionMethod.call(this, subject);
                    }
                }, _b.id = node.id || 'id', _b.displayName = node.name, _b.repository = node.repository, _b.permissionPropertyKey = node.permissionProperty || 'permissions', _b));
            }
            var _a, _b;
        };
        nodeList.forEach(createClass);
        return classGraph;
    }
    /**
     *  Build map of nodeName -> SubjectClass for gracl hierarchy
        NOTE: this is a duplicate of buildResourceHierachy that
              is necessary for reasons explained above due to limitations
              in the type system.
     */
    static buildSubjectHierachy(schemaNodes) {
        const nodeList = util_1.topologicalSort(schemaNodes),
              classGraph = new Map();
        const createClass = node => {
            const getParentsMethod = node.getParents || Node_1.Node.prototype.getParents;
            if (node.parent) {
                const ParentClass = classGraph.get(node.parent);
                classGraph.set(node.name, (_a = class extends ParentClass {
                    getParents() {
                        return getParentsMethod.call(this);
                    }
                }, _a.id = node.id || 'id', _a.parentId = node.parentId, _a.displayName = node.name, _a.repository = node.repository, _a.permissionPropertyKey = node.permissionProperty || 'permissions', _a));
            } else {
                classGraph.set(node.name, (_b = class extends Subject_1.Subject {}, _b.id = node.id || 'id', _b.displayName = node.name, _b.repository = node.repository, _b.permissionPropertyKey = node.permissionProperty || 'permissions', _b));
            }
            var _a, _b;
        };
        nodeList.forEach(createClass);
        return classGraph;
    }
    /**
     *  Extract a node class from the graph if it exists
     */
    getClass(name, type) {
        let map;
        switch (type) {
            case 'subject':
                map = this.subjects;
                break;
            case 'resource':
                map = this.resources;
                break;
            default:
                throw new Error(`Invalid class type ${ type }, must be subject or resource!`);
        }
        if (map.has(name)) return map.get(name);
        throw new Error(`No ${ type } class found for ${ name }!`);
    }
    /**
     *  Extract a resource class from the graph if it exists
     */
    getResource(name) {
        return this.getClass(name, 'resource');
    }
    /**
     *  Extract a subject class from the graph if it exists
     */
    getSubject(name) {
        return this.getClass(name, 'subject');
    }
}
exports.Graph = Graph;