"use strict";

var __awaiter = undefined && undefined.__awaiter || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) {
            try {
                step(generator.next(value));
            } catch (e) {
                reject(e);
            }
        }
        function rejected(value) {
            try {
                step(generator.throw(value));
            } catch (e) {
                reject(e);
            }
        }
        function step(result) {
            result.done ? resolve(result.value) : new P(function (resolve) {
                resolve(result.value);
            }).then(fulfilled, rejected);
        }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const util_1 = require('../util');
/**
 *  Abstract base class from which all gracl hierachy nodes inherit.
 */
class Node {
    /**
     *  Constructor, simply assigns the given document as a property
     */
    constructor(doc) {
        this.doc = doc;
        // ensure that this class has a repository

        var _getClass = this.getClass();

        const name = _getClass.name;
        const repository = _getClass.repository;
        const id = _getClass.id;

        if (!doc) throw new Error(`No document provided to ${ name } constructor!`);
        if (doc[id] === undefined) throw new Error(`No ${ id } property on document ${ doc }!`);
        if (!repository) throw new Error(`No repository static property defined on ${ name }!`);
    }
    /**
     *  Retrieve all class names in the hierarchy of nodes stemming from this class
     */
    static getHierarchyClassNames() {
        const names = [];
        let nodeClass = this;
        do {
            names.push(nodeClass.displayName || nodeClass.name);
            nodeClass = util_1.getClassOf(nodeClass.prototype);
        } while (util_1.getClassOf(nodeClass.prototype) !== Node);
        return names;
    }
    static getNodeDepth() {
        let depth = 0;
        let nodeClass = this;
        while (util_1.getClassOf(nodeClass.prototype) !== Node) {
            depth++;
            nodeClass = util_1.getClassOf(nodeClass.prototype);
        }
        return depth;
    }
    /**
     *  Get the name of this node
     */
    getName() {
        const thisClass = this.getClass(),
              className = thisClass.displayName || thisClass.name;
        return className;
    }
    /**
     *  Pretty printing
     */
    toString() {
        const nodeSubclassName = this.getNodeSubclass().name,
              id = this.getId();
        return `<${ nodeSubclassName }:${ this.getName() } id=${ id }>`;
    }
    /**
     *  Check if this node is a particular Node subclass
     */
    isNodeType(nc) {
        return this.getClass() === nc;
    }
    /**
     *  Get the relative super class constructor of this instance
     */
    getParentClass() {
        return util_1.getClassOf(this.constructor.prototype);
    }
    /**
     *  Get the class of this instance
     */
    getClass() {
        return util_1.getClassOf(this);
    }
    /**
     *  Check if this class direcly inherits from HierarchyNode by
        checking if the class two levels up is Node
     */
    hierarchyRoot() {
        return util_1.getClassOf(this.getParentClass().prototype) === Node;
    }
    /**
     *  Get the id value on the document contained in this node
     */
    getId() {
        return this.doc[this.getClass().id];
    }
    /**
     *  Get the repository for this class
     */
    getRepository() {
        return this.getClass().repository;
    }
    /**
     *  Check if a node is allowed access to this node. Must be overridden by subclasses.
     */
    isAllowed(node, permissionType, options) {
        return __awaiter(this, void 0, Promise, function* () {
            console.warn(`Calling Node.isAllowed(), must implement on subclass!`);
            return false;
        });
    }
    /**
     *  Get the parent objects of an instance of this node.
        Must be overriden by subclass unless the static parentId is defined.
     */
    getParents() {
        return __awaiter(this, void 0, Promise, function* () {
            var _getClass2 = this.getClass();

            const parentId = _getClass2.parentId;

            if (parentId) {
                const parentIds = this.doc[parentId] || [];
                if (Array.isArray(parentIds)) {
                    const promises = parentIds.map(id => {
                        return this.getParentNode(id);
                    });
                    const nodes = yield Promise.all(promises);
                    return Promise.resolve(nodes);
                } else {
                    return [yield this.getParentNode(parentIds)];
                }
            } else {
                console.warn(`Calling Node.getParents() without Node.parentId, must implement on subclass!`);
                return [];
            }
        });
    }

    /**
     *  Given an id of a parent of this node, create a Node instance of that object.
        @param data Either the <string> id of the object, or the raw document itself.
     */
    getParentNode(data) {
        return __awaiter(this, void 0, Promise, function* () {
            const ParentClass = this.getParentClass();
            let doc;
            // data is the id, retrieve from repository
            if (typeof data === 'string') {
                if (!ParentClass.repository) {
                    throw new Error(`No static repository property present on ${ ParentClass.name } Node!`);
                }
                doc = yield ParentClass.repository.getEntity(data, this);
            } else {
                doc = data;
            }
            return new ParentClass(doc);
        });
    }
    /**
     *  Ensure that a given class inherits from Node
     */
    assertNodeClass(nodeClass) {
        if (!(nodeClass.prototype instanceof Node)) {
            throw new Error(`Link in hierarchy chain (${ nodeClass.name }) is not an instance of Node!`);
        }
    }
    /**
     *  Determine what subclass of node this node is.
     */
    getNodeSubclass() {
        let nodeClass = this.getClass();
        this.assertNodeClass(nodeClass);
        while (util_1.getClassOf(nodeClass.prototype) !== Node) {
            nodeClass = util_1.getClassOf(nodeClass.prototype);
            this.assertNodeClass(nodeClass);
        }
        return nodeClass;
    }
    /**
     *  Determine what subclass of node this node is.
     */
    getNodeDepth() {
        return this.getClass().getNodeDepth();
    }
    /**
     *  Retrieve all ids in the hierarchy of nodes steming from this instance
     */
    getHierarchyIds() {
        return __awaiter(this, void 0, Promise, function* () {
            let ids = [this.getId()];
            if (!this.hierarchyRoot()) {
                const parents = yield this.getParents();
                if (parents.length) {
                    const parentIds = yield Promise.all(parents.map(p => p.getHierarchyIds()));
                    ids = parentIds.reduce((out, idList) => {
                        return out.concat(idList);
                    }, ids);
                }
            }
            return ids;
        });
    }
    /**
     *  Instance version of static method
     */
    getHierarchyClassNames() {
        return this.getClass().getHierarchyClassNames();
    }
}
Node.displayName = '';
Node.permissionPropertyKey = 'permissions';
/**
 *  The id property within the document contained by this node type, defaults to 'id'
 */
Node.id = 'id';
exports.Node = Node;