"use strict";

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i]; return arr2; } else { return Array.from(arr); } }

var __awaiter = undefined && undefined.__awaiter || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) {
            try {
                step(generator.next(value));
            } catch (e) {
                reject(e);
            }
        }
        function rejected(value) {
            try {
                step(generator.throw(value));
            } catch (e) {
                reject(e);
            }
        }
        function step(result) {
            result.done ? resolve(result.value) : new P(function (resolve) {
                resolve(result.value);
            }).then(fulfilled, rejected);
        }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const Node_1 = require('./Node');
const util_1 = require('../util');
/**
 * Resource class, permissions are stored on these Nodes, and Subject nodes
   check for access to these nodes.
 */
class Resource extends Node_1.Node {
    /**
     * Create new permissions property on document if not existing.
     */
    constructor(doc) {
        super(doc);
        const key = this.getClass().permissionPropertyKey;
        if (key && !this.doc[key]) {
            this.doc[key] = [];
        }
        this.sortPermissions();
    }
    /**
     * Sort the permissions on this document by subjectId to allow for fast searching.
     */
    sortPermissions() {
        const key = this.getClass().permissionPropertyKey;
        if (key && this.doc[key]) {
            const key = this.getClass().permissionPropertyKey;
            this.doc[key].sort(util_1.permissionCompare);
        }
        return this;
    }
    setDoc(doc) {
        this.doc = doc;
        const key = this.getClass().permissionPropertyKey;
        if (key && !this.doc[key]) {
            this.doc[key] = [];
            this.sortPermissions();
        }
        return this;
    }
    /**
     *  Retrieve a permission for a given subject via binary search.
        Returns an empty permission object if none is found.
     */
    getPermission(subject) {
        return __awaiter(this, void 0, Promise, function* () {
            const key = this.getClass().permissionPropertyKey,
                  subjectId = subject.getId(),
                  permissions = yield this.getPermissionList();
            return permissions[util_1.permissionIndexOf(permissions, subjectId)] || {
                subjectId: subjectId,
                resourceId: '',
                resourceType: '',
                subjectType: this.getName(),
                access: {}
            };
        });
    }
    getPermissionList() {
        return __awaiter(this, void 0, Promise, function* () {
            const key = this.getClass().permissionPropertyKey;
            return this.doc[key];
        });
    }
    /**
     *  Determine access to a given permission, provide a reason for access result.
           Steps for checking access for a given permission:
          1. If an assertion function is provided, make sure it returns true
              - return false if not
          2. Check if there is a permission on this resource specifically referencing the given subject
              - return true || false if the permission is not undefined
          3. Recurse up resource hierarchy, checking if parent resource has access to this resource.
              - if a parent subject specifically has true / false access -- return that boolean
          4. Recurse up subject hierarchy
     */
    determineAccess(subject, permissionType, options) {
        return __awaiter(this, void 0, Promise, function* () {
            // permission check options

            var _ref = options || {};

            var _ref$assertionFn = _ref.assertionFn;
            const assertionFn = _ref$assertionFn === undefined ? util_1.yes : _ref$assertionFn;

            const result = {
                type: permissionType,
                access: false,
                reason: 'No permissions were set specifically for this subject/resource combination.'
            };
            /**
             * Check if assertion function has returned true
             */
            if (!(yield assertionFn())) {
                result.access = false;
                result.reason = 'Failed assertion function check.';
                return result;
            }
            const subjectCache = {};
            const getSubject = sub => __awaiter(this, void 0, void 0, function* () {
                const id = sub.getId();
                if (subjectCache[id]) return subjectCache[id];
                return subjectCache[id] = yield sub.getParents();
            });
            const resource = this;
            let currentResources = [resource];
            // loop through the levels of the resource hierarchy
            while (true) {
                // loop through all levels of the subject hierarchy,
                // checking access to the current resource
                let currentSubjects = [subject];
                while (true) {
                    let accessSetAtCurrentLevel = false;
                    for (let i_s = 0, l_s = currentSubjects.length; i_s < l_s; i_s++) {
                        const sub = currentSubjects[i_s];
                        for (let i_r = 0, l_r = currentResources.length; i_r < l_r; i_r++) {
                            const res = currentResources[i_r];
                            // get the specific permission for this
                            // subject and resource combination and
                            // determine access for the given resource type
                            const access = (yield res.getPermission(sub)).access[permissionType];
                            // if we have a defined access value,
                            // set the reason and the access for the permission
                            if (access === true || access === false) {
                                accessSetAtCurrentLevel = true;
                                result.access = access;
                                result.reason = `Permission set on ${ res.toString() } for ${ sub.toString() } = ${ access }`;
                            }
                            // short circuit on false at this
                            // level of subjects and this level of resources
                            if (access === false) return result;
                        }
                    }
                    // if access was set to true for a given combination of
                    // subjects and resources, and no other combinations had false
                    // short circuit on true
                    if (accessSetAtCurrentLevel) return result;
                    // if we've reached the root, return
                    if (!currentSubjects.length || currentSubjects.some(s => s.hierarchyRoot())) break;
                    currentSubjects = yield util_1.flattenPromises(currentSubjects.map(getSubject));
                }
                if (!currentResources.length || currentResources.some(r => r.hierarchyRoot())) break;
                // advance to the next level of resources
                currentResources = yield util_1.flattenPromises(currentResources.map(r => r.getParents()));
            }
            return result;
        });
    }
    /**
     *  Check if a subject has access to this resource.
     */
    isAllowed(subject, permissionType, options) {
        return __awaiter(this, void 0, Promise, function* () {
            const result = yield this.determineAccess(subject, permissionType, options);
            return result.access;
        });
    }
    /**
     *  Get a string explaining why a subject has a permission set to a particular value for a given resource.
     */
    explainPermission(subject, permissionType, options) {
        return __awaiter(this, void 0, Promise, function* () {
            const result = yield this.determineAccess(subject, permissionType, options);
            return result.reason;
        });
    }
    /**
     *  Modify a permission on this Resource for a given Subject, using `action`
        to modify the permissions object. If no permission exists, a new permission is created.
     */
    updatePermission(subject, action) {
        return __awaiter(this, void 0, Promise, function* () {
            const permissions = yield this.getPermissionList(),
                  subjectId = subject.getId(),
                  subjectType = subject.getName(),
                  resourceId = this.getId(),
                  resourceType = this.getName();
            if (!permissions) {
                throw new Error(`No permissions available to update in updatePermission!`);
            }
            const existingPermissionIndex = util_1.permissionIndexOf(permissions, subjectId),
                  CurrentResourceClass = this.getClass();
            if (existingPermissionIndex >= 0) {
                permissions[existingPermissionIndex] = action(permissions[existingPermissionIndex]);
            } else {
                // add permission
                permissions.push(action({ subjectId: subjectId, resourceId: resourceId, resourceType: resourceType, subjectType: subjectType }));
            }
            // save updated document
            const id = this.getId(),
                  updated = yield CurrentResourceClass.repository.saveEntity(id, this.doc, this);
            return this.setDoc(updated);
        });
    }
    /**
     *  Set access for a particular permission type to true or false for a given Subject.
     */
    setPermissionAccess(subject, permissionType, access) {
        return this.updatePermission(subject, permission => {
            (permission.access = permission.access || {})[permissionType] = access;
            return permission;
        });
    }
    /**
     *  Set access for a particular permission type to true for a given Subject.
     */
    allow(subject, permissionType) {
        return this.setPermissionAccess(subject, permissionType, true);
    }
    /**
     *  Set access for a particular permission type to false for a given Subject.
     */
    deny(subject, permissionType) {
        return this.setPermissionAccess(subject, permissionType, false);
    }
    /**
     *  Retrieve permissions hierarchy for this node.
     */
    getPermissionsHierarchy() {
        return __awaiter(this, void 0, Promise, function* () {
            const permissions = yield this.getPermissionList();
            const graph = {
                node: this.toString(),
                nodeId: this.getId(),
                permissions: permissions,
                parents: []
            };
            if (!this.hierarchyRoot()) {
                const parents = yield this.getParents();
                if (parents.length) {
                    var _graph$parents;

                    const parentHierarchies = yield Promise.all(parents.map(p => p.getPermissionsHierarchy()));
                    (_graph$parents = graph.parents).push.apply(_graph$parents, _toConsumableArray(parentHierarchies));
                }
            }
            return graph;
        });
    }
}
exports.Resource = Resource;